from kawin.solver.Iterators import ExplicitEulerIterator, RK4Iterator
from enum import Enum
import time

class SolverType(Enum):
    EXPLICITEULER = 0
    RK4 = 1

class DESolver:
    '''
    Generic class for ODE/PDE solvers

    Generalization - coupled ODEs or PDEs (bunch of coupled ODEs) can be stated as dX/dt = f(X, t)

    Parameters
    ----------
    iterator : SolverType or Iterator
        Defines what iteration scheme to use
    defaultDt : float (defaults to 0.1)
        Default time increment if no function is implement to estimate a good time increment
    minDtFrac : float (defaults to 1e-8)
        Minimum time step as a fraction of simulation time
    maxDtFrac : float (defaults to 1)
        Maximum time step as a fraction of simulation time
    '''
    def __init__(self, iterator = SolverType.RK4, defaultDT = 0.1, minDtFrac = 1e-8, maxDtFrac = 1):
        self.dtmin = minDtFrac       #Min and max dt fraction of simulation time
        self.dtmax = maxDtFrac
        self.dt = defaultDT

        self.setFunctions(self.defaultPreProcess, self.defaultPostProcess, self.defaultPrintHeader, self.defaultPrintStatus)

        self.setIterator(iterator)

    def setIterator(self, iterator):
        '''
        Parameters
        ----------
        iterator : SolverType or Iterator
            Defines what iteration scheme to use
        '''
        if iterator == SolverType.EXPLICITEULER:
            self.iterator = ExplicitEulerIterator
        elif iterator == SolverType.RK4:
            self.iterator = RK4Iterator
        else:
            self.iterator = iterator

    def setFunctions(self, preProcess = None, postProcess = None, printHeader = None, printStatus = None):
        '''
        Sets functions before solving

        If any of these are not defined, then the corresponding function will be the default defined here
            Except for getDt (which returns defaultDt), the other functions will do nothing
        '''
        self.preProcess = self.preProcess if preProcess is None else preProcess
        self.postProcess = self.postProcess if postProcess is None else postProcess
        self.printHeader = self.printHeader if printHeader is None else printHeader
        self.printStatus = self.printStatus if printStatus is None else printStatus

    def setdXdtFunctions(self, f, correctdXdt, getDt, flattenX, unflattenX):
        self._f = f
        self._correctdXdt = correctdXdt
        self._getDt = getDt
        self._flattenX = flattenX
        self._unflattenX = unflattenX

    def defaultDtFunc(self, dXdt):
        '''
        Returns the default time increment
        '''
        return self.dt
    
    def defaultPreProcess(self):
        '''
        Default pre-processing function before an iteration
        '''
        return
    
    def defaultPostProcess(self, currTime, X_new):
        '''
        Default post-processing function after an iteration
        '''
        return X_new, False
    
    def defaultPrintHeader(self):
        '''
        Default print function before solving
        '''
        return
    
    def defaultPrintStatus(self, iteration, modeltime, simTimeElapsed):
        '''
        Default print function for when n iterations passed and verbose is true
        '''
        return
    
    def correctdXdtNotImplemented(self, dt, x, dXdt):
        '''
        Default function to correct dXdt
        '''
        pass

    def flattenXNotImplemented(self, X):
        '''
        Default flattenX function, which assumes X is in the correct format
        '''
        return X
    
    def unflattenXNotImplemented(self, X_flat, X_ref):
        '''
        Default unflattenX function which assumes X is in the correct format
        '''
        return X_flat
    
    def _getdXdt(self, t, x, getDt = False):
        '''
        Wrapper around getdXdt which will handle the following:
            Handle flattening/unfalttening the x and dx/dt arrays
            Calculate dt if not supplied

        The API for the iterator will be that all arrays are 1D np.arrays where operators will be trivial

        Parameters
        ----------
        t : float
            Time
        x : 1D np.array
            Model values
        getDt : bool
            Will calculate dt if True
        '''
        unflatX = self._unflattenX(x, self._X0)
        dXdt = self._f(t, unflatX)
        if getDt:
            dt = self._getDt(dXdt)
            dt = dt if dt > self._dtmin else self._dtmin
            dt = dt if dt < self._dtmax else self._dtmax
            return self._flattenX(dXdt), dt
        else:
            return self._flattenX(dXdt)
        
    def _updateX(self, x, dxdt, dt):
        '''
        Helper function that hides the correctdXdt function

        The API for the iterator will be that all arrays are 1D np.arrays where operators will be trivial

        Parameters
        ----------
        x : 1D np.array
            Model values
        dxdt : 1D np.array
            Derivatives at x
        dt : float
            Time step
        '''
        unflatdxdt = self._unflattenX(dxdt, self._X0)
        self._correctdXdt(dt, self._X0, unflatdxdt)
        return x + self._flattenX(unflatdxdt)*dt
    
    def solve(self, t0, X0, tf, verbose = False, vIt = 10):
        '''
        Solves dX/dt over a time increment
        This will be the main function that a model will use

        Steps during each iteration
            1. Print status if vIt iterations passed
            2. preProcess
            3. Iterate
            4. Update current time
            5. postProcess

        Parameters
        ----------
        f : function
            dX/dt - function taking in time and returning dX/dt
        t0 : float
            Starting time
        X0 : list of arrays
            X at time t
        tf : float
            Final time
        verbose: bool (defaults to False)
            Whether to print status
        vIt : integer (defaults to 10)
            Number of iterations to print status
        '''
        if verbose:
            self.printHeader()

        self._dtmin = self.dtmin * (tf - t0)
        self._dtmax = self.dtmax * (tf - t0)
        currTime = t0
        i = 0
        timeStart = time.time()
        stop = False
        while currTime < tf and not stop:
            if verbose and i % vIt == 0:
                timeFinish = time.time()
                self.printStatus(i, currTime, timeFinish - timeStart)

            self.preProcess()
            #Limit dtmax to remaining time if it's larger
            if self._dtmax > tf - currTime:
                self._dtmax = tf - currTime

            #Store X0 as a reference variable for _unflattenX
            #We have to do this per iteration since the shape of X0 can change during postProcess
            #    This is especially true for the population balance model with adaptive bins
            #The iterator also returns the flat array of X, so we need to unflatten it afterwards here
            self._X0 = X0
            X0_flat, dt = self.iterator(self._getdXdt, currTime, self._flattenX(X0), self._updateX)
            X0 = self._unflattenX(X0_flat, self._X0)
            
            currTime += dt
            X0, stop = self.postProcess(currTime, X0)
            i += 1

        if verbose:
            if stop:
                print('Stopping condition met. Ending simulation early.')
                
            timeFinish = time.time()
            self.printStatus(i, currTime, timeFinish - timeStart)



import numpy as np
import scipy.special as ssp

class EffectiveDiffusionFunctions:
    '''
    Stores variables and functions to calculate effective diffusion distance
    '''
    def __init__(self):
        self.setupInterpolation()

    def setupInterpolation(self, n = 250, lmin = -5, lmax = 1):
        '''
        Sets up interpolation points for effective diffusion distance

        Parameters
        ----------
        n : int
            Number of points to interpolate between, two endpoints will be added
            at supersaturation of 0 and 1
        lmin : float
            Minimum lambda value in log scale
        lmax : float
            Maximum lambda value in log scale
        '''
        lamInterp = np.logspace(lmin, lmax, n)
        self.ohmInterp = 2*lamInterp**2 - 2*lamInterp**3 * np.sqrt(np.pi) * np.exp(lamInterp**2) * ssp.erfc(lamInterp)
        self.effDiffInterp = self.ohmInterp / (2 * lamInterp**2)
        self.ohmInterp = np.concatenate(([0], self.ohmInterp, [1]))
        self.effDiffInterp = np.concatenate(([1], self.effDiffInterp, [0]))

    def effectiveDiffusionDistance(self, supersaturation):
        '''
        Effective diffusion distance given supersaturation
        This gives a constant to be multiplied by the particle radius
        
        The effective diffusion distance is given by
        eps = Q/2*lambda
        
        Where Q is the super saturation and lambda is given by
        Q = 2*lambda^2 - 2*lambda^3 * sqrt(pi) * exp(lambda^2) * (1 - erf(lambda))
        
        Parameters
        ----------
        supersaturation : float or array of floats
            (x - x_P) / (x_M - x_P) where x is matrix composition, x_M is parent composition and x_P is precipitate composition
        
        Returns
        -------
        Effective diffusion distance (eps) to be used as (eps*R) in the growth rate equation
        '''
        return np.interp(supersaturation, self.ohmInterp, self.effDiffInterp)

    def lambdaLow(self, supersaturation):
        '''
        Lambda when Q approaches 0
        This is done to prevent precision errors when multiplying exp*(1-erf)
        '''
        return np.sqrt(supersaturation / 2)

    def lambdaHigh(self, supersaturation):
        '''
        Lambda when Q approaches 1
        This is done to prevent precision errors when multiplying exp*(1-erf)
        '''
        return np.sqrt(3 / (2 * (1 - supersaturation)))

    def effectiveDiffusionDistanceApprox(self, supersaturation):
        '''
        Effective diffusion distance given supersaturation
        This gives a constant to be multiplied by the particle radius
        
        When Q approaches 1, this equation oscillates due to floating point errors
        Therefore, interpolation is performed between the two limits (Q -> 0 and Q -> 1)

        This may be removed in later versions
        
        Parameters
        ----------
        supersaturation : float or array of floats
            (x - x_P) / (x_M - x_P) where x is matrix composition, x_M is parent composition and x_P is precipitate composition
        
        Returns
        -------
        Effective diffusion distance (eps) to be used as (eps*R) in the growth rate equation
        '''
        
        #Interpolation constant
        a = 1.2
        
        if hasattr(supersaturation, '__len__'):
            diff = np.zeros(len(supersaturation))
            indices = (supersaturation >= 0) & (supersaturation < 1)
            diff[supersaturation >= 1] = 0
            diff[supersaturation < 0] = 1
            
            lam = (1 - supersaturation[indices]**a) * self.lambdaLow(supersaturation[indices]) + (supersaturation[indices]**a) * self.lambdaHigh(supersaturation[indices])
            diff[indices] = supersaturation[indices] / (2 * lam**2)
            
            return diff
            
        else:
            if supersaturation < 0:
                return 1
            
            if supersaturation >= 1:
                return 0
                
            lam = (1 - supersaturation**a) * self.lambdaLow(supersaturation) + (supersaturation**a) * self.lambdaHigh(supersaturation)

            return supersaturation / (2 * lam**2)
        
    def noDiffusionDistance(self, supersaturation):
        '''
        If neglecting the effective diffusion distance, return 1

        Parameters
        ----------
        supersaturation : float or array of floats
            (x - x_P) / (x_M - x_P) where x is matrix composition, x_M is parent composition and x_P is precipitate composition
        
        Returns
        -------
        1 or array of 1s
        '''
        if hasattr(supersaturation, '__len__'):
            return np.ones(len(supersaturation), dtype=np.float32)
        else:
            return 1


import numpy as np
from numpy.testing import assert_allclose
from kawin.precipitation import PopulationBalanceModel

#Set parameters for pbm. Default bins are increased here so that added bins should be 50
bins = 200
qBins = int(200/4)
minBins = 100
maxBins = 300
pbm = PopulationBalanceModel(1e-10, 1e-8, 200, 100, 300)

def test_reset():
    '''
    Resetting the model should bring it to the initialized parameters
    '''
    #Add and adjust pbm arbitrarily, changing these should not affect this test function
    pbm.addSizeClasses(150)
    pbm.PSD[-1] = 2
    pbm.adjustSizeClassesEuler(False)

    pbm.reset()
    assert(len(pbm.PSD) == bins and pbm.bins == len(pbm.PSD))
    assert(len(pbm.PSDbounds) == bins+1)
    assert(len(pbm.PSDsize) == bins)
    assert(pbm.PSDbounds[0] == 1e-10)
    assert(pbm.PSDbounds[-1] == 1e-8)

def test_addBins():
    '''
    If last bin is filled, then the number of bins added should be default bins/2
    '''
    pbm.PSD[-1] = 2
    finalLength = pbm.PSDbounds[-1] + qBins * (pbm.PSDbounds[1] - pbm.PSDbounds[0])
    pbm.adjustSizeClassesEuler(False)
    assert(len(pbm.PSD) == bins+qBins and pbm.bins == len(pbm.PSD))
    assert_allclose(pbm.PSDbounds[-1], finalLength, rtol=1e-6)
    assert_allclose(pbm.max, finalLength, rtol=1e-6)
    assert(len(pbm.PSDbounds) == bins+qBins+1)
    assert(len(pbm.PSDsize) == bins+qBins)
    assert_allclose(pbm.PSDsize[0], 0.5*(pbm.PSDbounds[0] + pbm.PSDbounds[1]), atol=0, rtol=1e-6)
    pbm.reset()

def test_increaseBinSize():
    '''
    If number of bins > maxBins, then increase bin size to keep range with bins = minBins
    '''
    pbm.addSizeClasses(int(0.9*bins))
    pbm.PSD[-1] = 2
    finalLength = pbm.PSDbounds[-1] + qBins*(pbm.PSDbounds[1] - pbm.PSDbounds[0])
    pbm.adjustSizeClassesEuler(False)
    assert(len(pbm.PSD) == minBins and pbm.bins == len(pbm.PSD))
    assert_allclose(pbm.PSDbounds[-1], finalLength, rtol=1e-6)
    assert_allclose(pbm.max, finalLength, rtol=1e-6)
    assert(len(pbm.PSDbounds) == minBins+1)
    assert(len(pbm.PSDsize) == minBins)
    assert_allclose(pbm.PSDsize[0], 0.5*(pbm.PSDbounds[0] + pbm.PSDbounds[1]), atol=0, rtol=1e-6)
    pbm.reset()

def test_decreaseBinSize():
    '''
    If max filled bin < 1/2 minBins, then decrease bin size so the last filled bin is the max and bins = maxBins
    '''
    filledBin = int(1/4 * minBins)
    pbm.PSD[filledBin] = 2
    finalLength = pbm.PSDbounds[filledBin+1]
    pbm.adjustSizeClassesEuler(True)
    assert(len(pbm.PSD) == maxBins and pbm.bins == len(pbm.PSD))
    assert_allclose(pbm.PSDbounds[-1], finalLength, rtol=1e-6)
    assert_allclose(pbm.max, finalLength, rtol=1e-6)
    assert(len(pbm.PSDbounds) == maxBins+1)
    assert(len(pbm.PSDsize) == maxBins)
    assert_allclose(pbm.PSDsize[0], 0.5*(pbm.PSDbounds[0] + pbm.PSDbounds[1]), atol=0, rtol=1e-6)
    pbm.reset()

def test_DT():
    '''
    Calculated DT with constant growth rate
    DT = ratio * binSize / (max(growth rate))

    Previous version had ratio of 0.5, but this was decreased slightly to 0.4 for numerical stability
    '''
    growth = 5*np.ones(pbm.bins+1)
    pbm.PSD = 2*np.ones(pbm.bins)
    ratio = 0.4
    trueDT = ratio * (pbm.PSDbounds[1] - pbm.PSDbounds[0]) / (growth[0])
    dissIndex = pbm.getDissolutionIndex(1e-3, 0)
    calcDT = pbm.getDTEuler(5, growth, dissIndex, maxBinRatio=ratio)
    assert_allclose(trueDT, calcDT, rtol=1e-6)



import numpy as np
import matplotlib.pyplot as plt
from kawin.precipitation import PrecipitateModel
from kawin.precipitation.StoppingConditions import PrecipitationStoppingCondition
from typing import List

class TTPCalculator:
    '''
    Time-temperature-precipitation

    Parameters
    ----------
    model : PrecipitateModel
    stopConds : list of PrecipitateStoppingConditions
        Stopping conditions to store times when these conditions are reached
        Model will continue to solve until the max time is reached or all conditions are satisfied
    '''
    def __init__(self, model : PrecipitateModel, stopConds : List[PrecipitationStoppingCondition]):
        self.model = model
        self.stopConds = stopConds
        self._maxTime = 0
        self.transformationTimes = None

        #Add stopping conditions to model
        #NOTE: this clears any previous stopping conditions
        self.model.clearStoppingConditions()
        for j in range(len(stopConds)):
            self.model.addStoppingCondition(self.stopConds[j], 'and')

    def _getStopTime(self, T):
        '''
        Internal function to get times for each stopping conditions at a single temperature

        Parameters
        ----------
        T : float
            Temperature
        '''
        self.model.reset()
        self.model.setTemperature(T)
        self.model.solve(self._maxTime, verbose = True, vIt = 1000)

        values = np.zeros(len(self.stopConds))
        for j in range(len(self.stopConds)):
            values[j] = self.stopConds[j].satisfiedTime()

        return values
    
    def calculateTTP(self, Tlow, Thigh, Tsteps, maxTime, pool = None):
        '''
        Calculates TTP diagram between Tlow and Thigh

        Parameters
        ----------
        Tlow : float
            Lower temperature range
        Thigh : float
            Upper temperature range
        Tsteps : int
            Number of temperatures between Tlow and Thigh to evaluate
        maxTime : float
            Maximum simulation time
            If the model reaches the max time before all stopping conditions are met, it will stop prematurely
                and any unsatisfied stopping conditions will be recorded as -1
        pool : None or multiprocessing pool
            If None, each temperature will be evaluated in serial
            If a pool, must have a map function
                Possible options: 
                    multiprocessing.Pool - (mac and unix only)
                    pathos.multiprocessing.ProcessingPool - (windows, mac and unix)
                    dask.Client - (windows, mac and unix)
        '''
        self.transformationTimes = np.zeros((Tsteps, len(self.stopConds)))
        self._maxTime = maxTime
        self.temperatures = np.linspace(Tlow, Thigh, Tsteps)

        if pool is None:
            outputs = list(map(self._getStopTime, self.temperatures))
        else:
            outputs = list(pool.map(self._getStopTime, self.temperatures))

        for i in range(len(self.temperatures)):
            for j in range(len(self.stopConds)):
                self.transformationTimes[i,j] = outputs[i][j]

    def plot(self, ax, labels, xlim = [1, 1e6], *args, **kwargs):
        '''
        Plots TTP diagram

        Parameters
        ----------
        ax : Matplotlib axes object
        labels : list of str
            Labels for each stopping condition
        xlim : list of float
            x-axis limits
            Plotting will be set on log scale, so lower limits will be set to be non-zero
        '''
        for i in range(len(self.stopConds)):
            indices = self.transformationTimes[:,i] != -1
            ax.plot(self.transformationTimes[indices,i], self.temperatures[indices], label=labels[i], *args, **kwargs)
        ax.legend()
        if xlim[0] == 0:
            xlim[0] = 1e-3
        ax.set_xlim(xlim)
        ax.set_xlabel('Time (s)')
        ax.set_xscale('log')
        ax.set_ylabel('Temperature (K)')


import numpy as np
import time
import csv
from itertools import zip_longest
from kawin.solver.Solver import DESolver, SolverType
from kawin.GenericModel import GenericModel
import kawin.diffusion.Plot as diffPlot

class DiffusionModel(GenericModel):
    #Boundary conditions
    FLUX = 0
    COMPOSITION = 1

    def __init__(self, zlim, N, elements = ['A', 'B'], phases = ['alpha'], record = True):
        '''
        Class for defining a 1-dimensional mesh

        Parameters
        ----------
        zlim : tuple
            Z-bounds of mesh (lower, upper)
        N : int
            Number of nodes
        elements : list of str
            Elements in system (first element will be assumed as the reference element)
        phases : list of str
            Number of phases in the system
        '''
        super().__init__()
        if isinstance(phases, str):
            phases = [phases]
        self.zlim, self.N = zlim, N
        self.allElements, self.elements = elements, elements[1:]
        self.phases = phases
        self.therm = None

        self.z = np.linspace(zlim[0], zlim[1], N)
        self.dz = self.z[1] - self.z[0]
        self.t = 0

        self.reset()

        self.LBC, self.RBC = self.FLUX*np.ones(len(self.elements)), self.FLUX*np.ones(len(self.elements))
        self.LBCvalue, self.RBCvalue = np.zeros(len(self.elements)), np.zeros(len(self.elements))

        self.cache = True
        self.setHashSensitivity(4)
        self.minComposition = 1e-8

        self.maxCompositionChange = 0.002

        if record:
            self.enableRecording()
        else:
            self.disableRecording()
            self._recordedX = None
            self._recordedP = None
            self._recordedZ = None
            self._recordedTime = None

    def reset(self):
        '''
        Resets model

        This involves clearing any caches in the Thermodynamics object and this model
        as well as resetting the composition and phase profiles
        '''
        if self.therm is not None:
            self.therm.clearCache()
        
        self.x = np.zeros((len(self.elements), self.N))
        self.p = np.ones((1,self.N)) if len(self.phases) == 1 else np.zeros((len(self.phases), self.N))
        self.hashTable = {}
        self.isSetup = False
        self.t = 0

    def setThermodynamics(self, thermodynamics):
        '''
        Defines thermodynamics object for the diffusion model

        Parameters
        ----------
        thermodynamics : Thermodynamics object
            Requires the elements in the Thermodynamics and DiffusionModel objects to have the same order
        '''
        self.therm = thermodynamics

    def setTemperature(self, T):
        '''
        Sets iso-thermal temperature

        Parameters
        ----------
        T : float
            Temperature in Kelvin
        '''
        self.Tparam = T
        self.T = T
        self.Tfunc = lambda z, t: self.Tparam * np.ones(len(z))

    def setTemperatureArray(self, times, temperatures):
        self.Tparam = (times, temperatures)
        self.T = temperatures[0]
        self.Tfunc = lambda z, t: np.interp(t/3600, self.Tparam[0], self.Tparam[1], self.Tparam[1][0], self.Tparam[1][-1]) * np.ones(len(z))

    def setTemperatureFunction(self, func):
        '''
        Function should be T = (x, t)
        '''
        self.Tparam = func
        self.Tfunc = lambda z, t: self.Tparam(z, t)

    def _getVarDict(self):
        '''
        Returns mapping of { variable name : attribute name } for saving
        The variable name will be the name in the .npz file
        '''
        saveDict = {
            'elements': 'elements',
            'phases': 'phases',
            'z': 'z',
            'zLim': 'zLim',
            'N': 'N',
            'finalTime': 't',
            'finalX': 'x',
            'finalP': 'p',
            'recordX': '_recordedX',
            'recordP': '_recordedP',
            'recordZ': '_recordedZ',
            'recordTime': '_recordedTime',
        }
        return saveDict

    def load(filename):
        '''
        Loads data from filename and returns a PrecipitateModel
        '''
        data = np.load(filename)
        model = DiffusionModel(data['zLim'], data['N'], data['elements'], data['phases'])
        model._loadData(data)
        model.isSetup = True
        return model

    def setHashSensitivity(self, s):
        '''
        Sets sensitivity of the hash table by significant digits

        For example, if a composition set is (0.5693, 0.2937) and s = 3, then
        the hash will be stored as (0.569, 0.294)

        Lower s values will give faster simulation times at the expense of accuracy

        Parameters
        ----------
        s : int
            Number of significant digits to keep for the hash table
        '''
        self.hashSensitivity = np.power(10, int(s))

    def _getHash(self, x, T):
        '''
        Gets hash value for a composition set

        Parameters
        ----------
        x : list of floats
            Composition set to create hash
        '''
        return hash(tuple((np.concatenate((x, [T]))*self.hashSensitivity).astype(np.int32)))

    def useCache(self, use):
        '''
        Whether to use the hash table

        Parameters
        ----------
        use : bool
            If True, then the hash table will be used
        '''
        self.cache = use

    def clearCache(self):
        '''
        Clears hash table
        '''
        self.hashTable = {}

    def enableRecording(self, dtype = np.float32):
        '''
        Enables recording of composition and phase
        
        Parameters
        ----------
        dtype : numpy data type (optional)
            Data type to record particle size distribution in
            Defaults to np.float32
        '''
        self._record = True
        self._recordedX = np.zeros((1, len(self.elements), self.N))
        self._recordedP = np.zeros((1, 1,self.N)) if len(self.phases) == 1 else np.zeros((1, len(self.phases), self.N))
        self._recordedZ = self.z
        self._recordedTime = np.zeros(1)

    def disableRecording(self):
        '''
        Disables recording
        '''
        self._record = False

    def removeRecordedData(self):
        '''
        Removes recorded data
        '''
        self._recordedX = None
        self._recordedP = None
        self._recordedZ = None
        self._recordedTime = None

    def record(self, time):
        '''
        Adds current mesh data to recorded arrays
        '''
        if self._record:
            if time > 0:
                self._recordedX = np.pad(self._recordedX, ((0, 1), (0, 0), (0, 0)))
                self._recordedP = np.pad(self._recordedP, ((0, 1), (0, 0), (0, 0)))
                self._recordedTime = np.pad(self._recordedTime, (0, 1))

            self._recordedX[-1] = self.x
            self._recordedP[-1] = self.p
            self._recordedTime[-1] = time

    def setMeshtoRecordedTime(self, time):
        '''
        From recorded values, interpolated at time to get composition and phase fraction
        '''
        if self._record:
            if time < self._recordedTime[0]:
                print('Input time is lower than smallest recorded time, setting PSD to t = {:.3e}'.format(self._recordedTime[0]))
                self.x, self.p = self._recordedX[0], self._recordedP[0]
            elif time > self._recordedTime[-1]:
                print('Input time is larger than longest recorded time, setting PSD to t = {:.3e}'.format(self._recordedTime[-1]))
                self.x, self.p = self._recordedX[-1], self._recordedP[-1]
            else:
                uind = np.argmax(self._recordedTime > time)
                lind = uind - 1

                ux, up, utime = self._recordedX[uind], self._recordedP[uind], self._recordedTime[uind]
                lx, lp, ltime = self._recordedX[lind], self._recordedP[lind], self._recordedTime[lind]

                self.x = (ux - lx) * (time - ltime) / (utime - ltime) + lx
                self.p = (up - lp) * (time - ltime) / (utime - ltime) + lp
            
            self.z = self._recordedZ

    def _getElementIndex(self, element = None):
        '''
        Gets index of element in self.elements

        Parameters
        ----------
        element : str
            Specified element, will return first element if None
        '''
        if element is None:
            return 0
        else:
            return self.elements.index(element)

    def _getPhaseIndex(self, phase = None):
        '''
        Gets index of phase in self.phases

        Parameters
        ----------
        phase : str
            Specified phase, will return first phase if None
        '''
        if phase is None:
            return 0
        else:
            return self.phases.index(phase)

    def setBC(self, LBCtype = 0, LBCvalue = 0, RBCtype = 0, RBCvalue = 0, element = None):
        '''
        Set boundary conditions

        Parameters
        ----------
        LBCtype : int
            Left boundary condition type
                Mesh1D.FLUX - constant flux
                Mesh1D.COMPOSITION - constant composition
        LBCvalue : float
            Value of left boundary condition
        RBCtype : int
            Right boundary condition type
                Mesh1D.FLUX - constant flux
                Mesh1D.COMPOSITION - constant composition
        RBCvalue : float
            Value of right boundary condition
        element : str
            Specified element to apply boundary conditions on
        '''
        eIndex = self._getElementIndex(element)
        self.LBC[eIndex] = LBCtype
        self.LBCvalue[eIndex] = LBCvalue
        if LBCtype == self.COMPOSITION:
            self.x[eIndex,0] = LBCvalue

        self.RBC[eIndex] = RBCtype
        self.RBCvalue[eIndex] = RBCvalue
        if RBCtype == self.COMPOSITION:
            self.x[eIndex,-1] = RBCvalue

    def setCompositionLinear(self, Lvalue, Rvalue, element = None):
        '''
        Sets composition as a linear function between ends of the mesh

        Parameters
        ----------
        Lvalue : float
            Value at left boundary
        Rvalue : float
            Value at right boundary
        element : str
            Element to apply composition profile to
        '''
        eIndex = self._getElementIndex(element)
        self.x[eIndex] = np.linspace(Lvalue, Rvalue, self.N)

    def setCompositionStep(self, Lvalue, Rvalue, z, element = None):
        '''
        Sets composition as a step-wise function

        Parameters
        ----------
        Lvalue : float
            Value on left side of mesh
        Rvalue : float
            Value on right side of mesh
        z : float
            Position on mesh where composition switches from Lvalue to Rvalue
        element : str
            Element to apply composition profile to
        '''
        eIndex = self._getElementIndex(element)
        Lindices = self.z <= z
        self.x[eIndex,Lindices] = Lvalue
        self.x[eIndex,~Lindices] = Rvalue

    def setCompositionSingle(self, value, z, element = None):
        '''
        Sets single node to specified composition

        Parameters
        ----------
        value : float
            Composition
        z : float
            Position to set value to (will use closest node to z)
        element : str
            Element to apply composition profile to
        '''
        eIndex = self._getElementIndex(element)
        zIndex = np.argmin(np.abs(self.z-z))
        self.x[eIndex,zIndex] = value

    def setCompositionInBounds(self, value, Lbound, Rbound, element = None):
        '''
        Sets single node to specified composition

        Parameters
        ----------
        value : float
            Composition
        Lbound : float
            Position of left bound
        Rbound : float
            Position of right bound
        element : str
            Element to apply composition profile to
        '''
        eIndex = self._getElementIndex(element)
        indices = (self.z >= Lbound) & (self.z <= Rbound)
        self.x[eIndex,indices] = value

    def setCompositionFunction(self, func, element = None):
        '''
        Sets composition as a function of z

        Parameters
        ----------
        func : function
            Function taking in z and returning composition
        element : str
            Element to apply composition profile to
        '''
        eIndex = self._getElementIndex(element)
        self.x[eIndex,:] = func(self.z)

    def setCompositionProfile(self, z, x, element = None):
        '''
        Sets composition profile by linear interpolation

        Parameters
        ----------
        z : array
            z-coords of composition profile
        x : array
            Composition profile
        element : str
            Element to apply composition profile to
        '''
        eIndex = self._getElementIndex(element)
        z = np.array(z)
        x = np.array(x)
        sortIndices = np.argsort(z)
        z = z[sortIndices]
        x = x[sortIndices]
        self.x[eIndex,:] = np.interp(self.z, z, x)

    def setup(self):
        '''
        General setup function for all diffusio models

        This will clear any cached values in the thermodynamics function and check if all compositions add up to 1

        This will also make sure that all compositions are not 0 or 1 to speed up equilibrium calculations
        '''
        if self.therm is not None:
            self.therm.clearCache()
        xsum = np.sum(self.x, axis=0)
        if any(xsum > 1):
            print('Compositions add up to above 1 between z = [{:.3e}, {:.3e}]'.format(np.amin(self.z[xsum>1]), np.amax(self.z[xsum>1])))
            raise Exception('Some compositions sum up to above 1')
        self.x[self.x > self.minComposition] = self.x[self.x > self.minComposition] - len(self.allElements) * self.minComposition
        self.x[self.x < self.minComposition] = self.minComposition
        self.T = self.Tfunc(self.z, 0)
        self.isSetup = True
        self.record(self.t) #Record at t = 0

    def _getFluxes(self):
        '''
        "Virtual" function to be implemented by child objects

        Should return (fluxes (list), dt (float))
        '''
        raise NotImplementedError()
    
    def printHeader(self):
        print('Iteration\tSim Time (h)\tRun time (s)')

    def printStatus(self, iteration, modelTime, simTimeElapsed):
        super().printStatus(iteration, modelTime/3600, simTimeElapsed)

    def getCurrentX(self):
        return self.t, [self.x]
    
    def getdXdt(self, t, x):
        '''
        dXdt is defined as -dJ/dz
        '''
        fluxes = self._getFluxes(t, x)
        return [-(fluxes[:,1:] - fluxes[:,:-1])/self.dz]
    
    def preProcess(self):
        return
    
    def postProcess(self, time, x):
        '''
        Stores new x and t
        Records new values if recording is enabled
        '''
        self.t = time
        self.x = x[0]
        self.record(self.t)
        self.updateCoupledModels()
        return self.getCurrentX()[1], False
    
    def flattenX(self, X):
        '''
        np.hstack does not flatten a 2D array, so we have to overload this function
            By itself, this doesn't actually affect the solver/iterator, but when coupled with other models,
            it becomes an issue

        This will convert the 2D array X to a 1D array by reshaping to 1D array of len(# elements * # nodes)
        '''
        return np.reshape(X[0], (np.prod(X[0].shape)))
    
    def unflattenX(self, X_flat, X_ref):
        '''
        Reshape X_flat to original shape
        '''
        return [np.reshape(X_flat, X_ref[0].shape)]

    def getX(self, element):
        '''
        Gets composition profile of element
        
        Parameters
        ----------
        element : str
            Element to get profile of
        '''
        if element in self.allElements and element not in self.elements:
            return 1 - np.sum(self.x, axis=0)
        else:
            e = self._getElementIndex(element)
            return self.x[e]

    def getP(self, phase):
        '''
        Gets phase profile

        Parameters
        ----------
        phase : str
            Phase to get profile of
        '''
        p = self._getPhaseIndex(phase)
        return self.p[p]

    def plot(self, ax = None, plotReference = True, plotElement = None, zScale = 1, *args, **kwargs):
        '''
        Plots composition profile

        Parameters
        ----------
        ax : matplotlib Axes object
            Axis to plot on
        plotReference : bool
            Whether to plot reference element (composition = 1 - sum(composition of rest of elements))
        plotElement : None or str
            Plots single element if it is defined, otherwise, all elements are plotted
        zScale : float
            Scale factor for z-coordinates
        '''
        return diffPlot.plot(self, ax, plotReference, plotElement, zScale, *args, **kwargs)

    def plotTwoAxis(self, Lelements, Relements, zScale = 1, axL = None, axR = None, *args, **kwargs):
        '''
        Plots composition profile with two y-axes

        Parameters
        ----------
        axL : matplotlib Axes object
            Left axis to plot on
        Lelements : list of str
            Elements to plot on left axis
        Relements : list of str
            Elements to plot on right axis
        axR : matplotlib Axes object (optional)
            Right axis to plot on
            If None, then the right axis will be created
        zScale : float
            Scale factor for z-coordinates
        '''
        return diffPlot.plotTwoAxis(self, Lelements, Relements, zScale, axL, axR, *args, **kwargs)

    def plotPhases(self, ax = None, plotPhase = None, zScale = 1, *args, **kwargs):
        '''
        Plots phase fractions over z

        Parameters
        ----------
        ax : matplotlib Axes object
            Axis to plot on
        plotPhase : None or str
            Plots single phase if it is defined, otherwise, all phases are plotted
        zScale : float
            Scale factor for z-coordinates
        '''
        return diffPlot.plotPhases(self, ax, plotPhase, zScale, *args, **kwargs)



from numpy.testing import assert_allclose
import numpy as np
from kawin.diffusion import SinglePhaseModel, HomogenizationModel
from kawin.thermo import GeneralThermodynamics
from kawin.tests.datasets import *

N = 100
singleModelBinary = SinglePhaseModel([-1e-3, 1e-3], N, ['NI', 'CR'], ['FCC_A1'])
singleModelTernary = SinglePhaseModel([-1e-3, 1e-3], N, ['NI', 'CR', 'AL'], ['FCC_A1'])
homogenizationBinary = HomogenizationModel([-1e-3, 1e-3], N, ['NI', 'CR'], ['FCC_A1', 'BCC_A2'])
homogenizationTernary = HomogenizationModel([-1e-3, 1e-3], N, ['NI', 'CR', 'AL'], ['FCC_A1', 'BCC_A2'])
NiCrTherm = GeneralThermodynamics(NICRAL_TDB, ['NI', 'CR'], ['FCC_A1', 'BCC_A2'])
NiCrAlTherm = GeneralThermodynamics(NICRAL_TDB, ['NI', 'CR', 'AL'], ['FCC_A1', 'BCC_A2'])
FeCrNiTherm = GeneralThermodynamics(FECRNI_DB, ['FE', 'CR', 'NI'], ['FCC_A1', 'BCC_A2'])

def test_CompositionInput():
    '''
    Tests that after setting up a model, all components are greater than 0
    
    In practice, this greatly speeds up simulation time without sacrificing accuracy since it avoids 
    performing equilibrium calculations with a composition of 0 for any given component

    The composition and setup functions for both models inherit from the
    base diffusion model class, so any model can be used here
    '''
    singleModelTernary.reset()
    singleModelTernary.setCompositionStep(0.2, 1, 0, 'CR')
    singleModelTernary.setCompositionStep(0.8, 0, 0, 'AL')
    singleModelTernary.setThermodynamics(NiCrAlTherm)
    singleModelTernary.setTemperature(1200+273.15)
    singleModelTernary.setup()

    assert(singleModelTernary.x[0,25] + singleModelTernary.x[1,25] < 1)
    assert(singleModelTernary.x[0,75] + singleModelTernary.x[1,75] < 1)
    assert(1 - (singleModelTernary.x[0,75] + singleModelTernary.x[1,75]) >= singleModelTernary.minComposition)
    assert(1 - (singleModelTernary.x[0,75] + singleModelTernary.x[1,75]) >= singleModelTernary.minComposition)
    assert(singleModelTernary.x[1,75] >= singleModelTernary.minComposition)

def test_SinglePhaseFluxes_shape():
    '''
    Tests the dimensions of the single phase fluxes function

    Should be (E-1, N+1) where E is the number of elements and
    N is the number of points
    '''
    singleModelBinary.reset()
    singleModelBinary.setCompositionStep(0.2, 0.8, 0, 'CR')
    singleModelBinary.setThermodynamics(NiCrTherm)
    singleModelBinary.setTemperature(1073)
    singleModelBinary.setup()

    fBinary, _ = singleModelBinary.getFluxes()

    singleModelTernary.reset()
    singleModelTernary.setCompositionStep(0.2, 0.4, 0, 'CR')
    singleModelTernary.setCompositionStep(0.4, 0.4, 0, 'AL')
    singleModelTernary.setThermodynamics(NiCrAlTherm)
    singleModelTernary.setTemperature(1073)
    singleModelTernary.setup()

    fTernary, _ = singleModelTernary.getFluxes()

    assert(fBinary.shape == (1,N+1))
    assert(fTernary.shape == (2,N+1))

def test_HomogenizationMobility():
    '''
    Tests the dimensions of the homogenization mobility function

    Should be (P, E, N) where P is the number of phases,
    E is the number of elements and
    N is the number of points
    '''
    homogenizationBinary.reset()
    homogenizationBinary.setCompositionStep(0.2, 0.8, 0, 'CR')
    homogenizationBinary.setThermodynamics(NiCrTherm)
    homogenizationBinary.setTemperature(1073)
    homogenizationBinary.setup()

    mobBinary = homogenizationBinary.getMobility(homogenizationBinary.x)

    homogenizationTernary.reset()
    homogenizationTernary.setCompositionStep(0.2, 0.4, 0, 'CR')
    homogenizationTernary.setCompositionStep(0.4, 0.4, 0, 'AL')
    homogenizationTernary.setThermodynamics(NiCrAlTherm)
    homogenizationTernary.setTemperature(1073)
    homogenizationTernary.setup()

    mobTernary = homogenizationTernary.getMobility(homogenizationTernary.x)

    assert(mobBinary.shape == (len(homogenizationBinary.phases),2,N))
    assert(mobTernary.shape == (len(homogenizationTernary.phases),3,N))

def test_homogenizationSinglePhaseMobility():
    '''
    Tests that in a single phase region, any of the mobility functions will give 
    the same mobility of the single phase itself
    '''
    homogenizationTernary.reset()
    #Ni-5Cr-5Al should always be FCC_A1
    homogenizationTernary.setCompositionStep(0.05, 0.4, 0, 'CR')
    homogenizationTernary.setCompositionStep(0.05, 0.4, 0, 'AL')
    homogenizationTernary.setThermodynamics(NiCrAlTherm)
    homogenizationTernary.setTemperature(1073)
    homogenizationTernary.setup()

    mob = homogenizationTernary.getMobility(homogenizationTernary.x)

    mobFuncs = ['wiener upper', 'wiener lower', 'hashin upper', 'hashin lower', 'lab']
    mobs = []
    for f in mobFuncs:
        homogenizationTernary.clearCache()
        homogenizationTernary.setup()
        homogenizationTernary.setMobilityFunction(f)
        mobs.append(homogenizationTernary.mobilityFunction(homogenizationTernary.x))
        assert(np.allclose(mobs[-1][:,0], mob[0,:,0], atol=0, rtol=1e-3))

def test_homogenization_wiener_upper():
    '''
    Tests output of wiener upper bounds in single and two-phase regions
    '''
    homogenizationTernary.clearCache()
    homogenizationTernary.reset()
    #Ni-5Cr-5Al should always be FCC_A1
    homogenizationTernary.setCompositionStep(0.05, 0.7, 0, 'CR')
    homogenizationTernary.setCompositionStep(0.05, 0.05, 0, 'AL')
    homogenizationTernary.setThermodynamics(NiCrAlTherm)
    homogenizationTernary.setTemperature(1073)
    homogenizationTernary.setup()
    homogenizationTernary.setMobilityFunction('wiener upper')

    mob = homogenizationTernary.mobilityFunction(homogenizationTernary.x)
    assert(np.allclose(mob[:,0], [3.927302e-22, 2.323337e-23, 6.206029e-23], atol=0, rtol=1e-3))
    assert(np.allclose(mob[:,-1], [2.025338e-22, 5.106062e-22, 8.524977e-23], atol=0, rtol=1e-3))

def test_homogenization_wiener_lower():
    '''
    Tests output of wiener upper bounds in single and two-phase regions
    '''
    homogenizationTernary.clearCache()
    homogenizationTernary.reset()
    #Ni-5Cr-5Al should always be FCC_A1
    homogenizationTernary.setCompositionStep(0.05, 0.7, 0, 'CR')
    homogenizationTernary.setCompositionStep(0.05, 0.05, 0, 'AL')
    homogenizationTernary.setThermodynamics(NiCrAlTherm)
    homogenizationTernary.setTemperature(1073)
    homogenizationTernary.setup()
    homogenizationTernary.setMobilityFunction('wiener lower')

    mob = homogenizationTernary.mobilityFunction(homogenizationTernary.x)
    assert(np.allclose(mob[:,0], [3.927302e-22, 2.323337e-23, 6.206029e-23], atol=0, rtol=1e-3))
    assert(np.allclose(mob[:,-1], [1.527894e-21, 3.851959e-21, 6.431152e-22], atol=0, rtol=1e-3))

def test_homogenization_hashin_upper():
    '''
    Tests output of wiener upper bounds in single and two-phase regions
    '''
    homogenizationTernary.clearCache()
    homogenizationTernary.reset()
    #Ni-5Cr-5Al should always be FCC_A1
    homogenizationTernary.setCompositionStep(0.05, 0.7, 0, 'CR')
    homogenizationTernary.setCompositionStep(0.05, 0.05, 0, 'AL')
    homogenizationTernary.setThermodynamics(NiCrAlTherm)
    homogenizationTernary.setTemperature(1073)
    homogenizationTernary.setup()
    homogenizationTernary.setMobilityFunction('hashin upper')

    mob = homogenizationTernary.mobilityFunction(homogenizationTernary.x)
    assert(np.allclose(mob[:,0], [3.927302e-22, 2.323337e-23, 6.206029e-23], atol=0, rtol=1e-3))
    assert(np.allclose(mob[:,-1], [1.536725e-22, 3.874223e-22, 6.468323e-23], atol=0, rtol=1e-3))

def test_homogenization_hashin_lower():
    '''
    Tests output of wiener upper bounds in single and two-phase regions
    '''
    homogenizationTernary.clearCache()
    homogenizationTernary.reset()
    #Ni-5Cr-5Al should always be FCC_A1
    homogenizationTernary.setCompositionStep(0.05, 0.7, 0, 'CR')
    homogenizationTernary.setCompositionStep(0.05, 0.05, 0, 'AL')
    homogenizationTernary.setThermodynamics(NiCrAlTherm)
    homogenizationTernary.setTemperature(1073)
    homogenizationTernary.setup()
    homogenizationTernary.setMobilityFunction('hashin lower')

    mob = homogenizationTernary.mobilityFunction(homogenizationTernary.x)
    assert(np.allclose(mob[:,0], [3.927302e-22, 2.323337e-23, 6.206029e-23], atol=0, rtol=1e-3))
    assert(np.allclose(mob[:,-1], [3.471117e-21, 8.751001e-21, 1.461049e-21], atol=0, rtol=1e-3))

def test_homogenization_lab():
    '''
    Tests output of wiener upper bounds in single and two-phase regions
    '''
    homogenizationTernary.clearCache()
    homogenizationTernary.reset()
    #Ni-5Cr-5Al should always be FCC_A1
    homogenizationTernary.setCompositionStep(0.05, 0.7, 0, 'CR')
    homogenizationTernary.setCompositionStep(0.05, 0.05, 0, 'AL')
    homogenizationTernary.setThermodynamics(NiCrAlTherm)
    homogenizationTernary.setTemperature(1073)
    homogenizationTernary.setup()
    homogenizationTernary.setMobilityFunction('lab')

    mob = homogenizationTernary.mobilityFunction(homogenizationTernary.x)
    assert(np.allclose(mob[:,0], [3.927302e-22, 2.323337e-23, 6.206029e-23], atol=0, rtol=1e-3))
    assert(np.allclose(mob[:,-1], [2.025338e-22, 5.106062e-22, 8.524977e-23], atol=0, rtol=1e-3))

def test_single_phase_dxdt():
    '''
    Check dxdt values of arbitrary single phase model problem

    We spot check a few points on dxdt rather than checking the entire array

    This uses the parameters from 06_Single_Phase_Diffusion example with the composition
    being linear rather then step functions
    '''
    #Define mesh spanning between -1mm to 1mm with 50 volume elements
    #Since we defined L12, the disordered phase as DIS_ attached to the front
    m = SinglePhaseModel([-1e-3, 1e-3], 20, ['NI', 'CR', 'AL'], ['FCC_A1'])

    #Define Cr and Al composition, with step-wise change at z=0
    m.setCompositionLinear(0.077, 0.359, 'CR')
    m.setCompositionLinear(0.054, 0.062, 'AL')

    m.setThermodynamics(NiCrAlTherm)
    m.setTemperature(1200 + 273.15)

    m.setup()
    t, x = m.getCurrentX()
    dxdt = m.getdXdt(t, x)
    dt = m.getDt(dxdt)

    #Index 5
    ind5, vals5 = 5, np.array([1.640437e-9, 5.669268e-10])

    #Index 10
    ind10, vals10 = 10, np.array([1.542640e-9, 1.091229e-9])

    #Index 15
    ind15, vals15 = 15, np.array([1.596203e-9, 1.842238e-9])
    
    assert_allclose(dxdt[0][:,ind5], vals5, atol=0, rtol=1e-3)
    assert_allclose(dxdt[0][:,ind10], vals10, atol=0, rtol=1e-3)
    assert_allclose(dxdt[0][:,ind15], vals15, atol=0, rtol=1e-3)
    assert_allclose(dt, 28721.530474, rtol=1e-3)

def test_diffusion_x_shape():
    '''
    Check the flatten and unflatten behavior for Diffusion model

    SinglePhaseModel and Homogenization model follows the same path for these functions
    since we just deal with fluxes for elements

    For this setup:
        getCurrentX will return a single element array with the element having a shape of (2,20)
        flattenX will return a 1D array of length 40 (2x20)
        unflattenX should take the output of flattenX and getCurrentX to bring the (40,) array to [(2,20)]
    '''
    #Define mesh spanning between -1mm to 1mm with 50 volume elements
    #Since we defined L12, the disordered phase as DIS_ attached to the front
    m = SinglePhaseModel([-1e-3, 1e-3], 20, ['NI', 'CR', 'AL'], ['DIS_FCC_A1'])

    #Define Cr and Al composition, with step-wise change at z=0
    m.setCompositionLinear(0.077, 0.359, 'CR')
    m.setCompositionLinear(0.054, 0.062, 'AL')

    m.setThermodynamics(NiCrAlTherm)
    m.setTemperature(1200 + 273.15)

    m.setup()
    t, x = m.getCurrentX()
    origShape = x[0].shape
    
    x_flat = m.flattenX(x)
    flatShape = x_flat.shape

    x_restore = m.unflattenX(x_flat, x)
    unflatShape = x_restore[0].shape

    assert(len(x) == 1)
    assert(origShape == unflatShape)
    assert(flatShape == (np.prod(origShape),))
    assert(len(x_restore) == 1)

def test_homogenization_dxdt():
    '''
    Check flux values of arbitrary homogenization model problem

    We spot check a few points on dxdt rather than checking the entire array
    
    We'll only test using the hashin lower homogenization function since there's already tests for 
    the output of each homogenization function

    This uses the parameters from 07_Homogenization_Model example with the compositions
    being linear rather than stepwise functions
    '''
    m = HomogenizationModel([-5e-4, 5e-4], 20, ['FE', 'CR', 'NI'], ['FCC_A1', 'BCC_A2'])
    m.setCompositionLinear(0.257, 0.423, 'CR')
    m.setCompositionLinear(0.065, 0.276, 'NI')
    m.setTemperature(1100+273.15)
    m.setThermodynamics(FeCrNiTherm)
    m.eps = 0.01

    m.setMobilityFunction('hashin lower')

    m.setup()
    t, x = m.getCurrentX()
    dxdt = m.getdXdt(t, x)
    dt = m.getDt(dxdt)
    
    #Index 5
    ind5, vals5 = 5, np.array([-1.581478e-9, 1.212876e-9])

    #Index 10
    ind10, vals10 = 10, np.array([-9.722631e-10, 1.703447e-9])

    #Index 15
    ind15, vals15 = 15, np.array([-4.720562e-10, 8.600518e-10])
    
    assert_allclose(dxdt[0][:,ind5], vals5, atol=0, rtol=1e-3)
    assert_allclose(dxdt[0][:,ind10], vals10, atol=0, rtol=1e-3)
    assert_allclose(dxdt[0][:,ind15], vals15, atol=0, rtol=1e-3)
    assert_allclose(dt, 62271.050081, rtol=1e-3)



